import pytest
import math


def test_connector_type(default_gen_sine):
    assert default_gen_sine.connector_type in default_gen_sine.CONNECTOR_TYPES


def test_is_differential(default_gen_sine):
    assert type(default_gen_sine.is_differential) is bool


def test_impedance(default_gen_sine):
    assert type(default_gen_sine.impedance) is float
    assert default_gen_sine.impedance > 0


def test_resolution(default_gen_sine):
    assert type(default_gen_sine.resolution) is int
    assert default_gen_sine.resolution > 0


def test_out_min(default_gen_sine):
    assert type(default_gen_sine.out_min) is float
    assert default_gen_sine.out_min < 0


def test_out_max(default_gen_sine):
    assert type(default_gen_sine.out_max) is float
    assert default_gen_sine.out_max > 0


def test_is_controllable(default_gen_sine):
    assert default_gen_sine.is_controllable is True


def test_status(default_gen_sine):
    assert default_gen_sine.status in default_gen_sine.GENERATOR_STATUSES


def test_is_out_on(default_gen_sine):
    # Test getter
    assert type(default_gen_sine.is_out_on) is bool

    # Test setter
    for value in [True, False]:
        default_gen_sine.is_out_on = value
        assert default_gen_sine.is_out_on is value


def test_is_out_inv_available(default_gen_sine):
    # Test type
    assert type(default_gen_sine.is_out_inv_available) is bool
    # Test value
    assert default_gen_sine.is_out_inv_available is True


def test_is_out_inv(default_gen_sine):
    # Test getter
    assert type(default_gen_sine.is_out_inv) is bool

    # Test setter
    for value in [True, False]:
        default_gen_sine.is_out_inv = value
        assert default_gen_sine.is_out_inv is value


def test_start(default_gen_sine):
    assert default_gen_sine.start() is True


def test_stop(default_gen_sine):
    default_gen_sine.start()
    assert default_gen_sine.stop() is True


def test_is_running(default_gen_sine):
    default_gen_sine.start()
    assert default_gen_sine.is_running is True


def test_signal_types_available(default_gen_sine):
    assert type(default_gen_sine.signal_types_available) is tuple
    for sig_type in default_gen_sine.signal_types_available:
        assert sig_type in default_gen_sine.SIGNAL_TYPES


def test_signal_type(default_gen_sine):
    # Test getter
    assert default_gen_sine.signal_type in default_gen_sine.SIGNAL_TYPES

    # Test setter
    for sig_type in default_gen_sine.signal_types_available:
        default_gen_sine.signal_type = sig_type
        assert default_gen_sine.signal_type == sig_type


def test_amplitude_min(default_gen_sine):
    assert type(default_gen_sine.amplitude_min) is float
    assert default_gen_sine.amplitude_min < default_gen_sine.amplitude_max


def test_amplitude_max(default_gen_sine):
    assert type(default_gen_sine.amplitude_max) is float
    assert default_gen_sine.amplitude_max > default_gen_sine.amplitude_min


def test_amplitude(default_gen_sine):
    # Test getter
    assert type(default_gen_sine.amplitude) is float

    # Test setter
    for ampl in [default_gen_sine.amplitude_min,
                 (default_gen_sine.amplitude_max-default_gen_sine.amplitude_min)/2,
                 default_gen_sine.amplitude_max]:
        default_gen_sine.amplitude = ampl
        assert default_gen_sine.amplitude == ampl


def test_amplitude_ranges_available(default_gen_sine):
    assert type(default_gen_sine.amplitude_ranges_available) is tuple
    for ampl_range in default_gen_sine.amplitude_ranges_available:
        assert type(ampl_range) is float
        assert ampl_range > 0


def test_amplitude_range(default_gen_sine):
    # Test getter
    assert type(default_gen_sine.amplitude_range) is float
    assert default_gen_sine.amplitude_range > 0

    # Test setter
    for ampl_range in default_gen_sine.amplitude_ranges_available:
        default_gen_sine.amplitude_range = ampl_range
        assert default_gen_sine.amplitude_range == ampl_range


def test_is_amplitude_autorange(default_gen_sine):
    assert type(default_gen_sine.is_amplitude_autorange) is bool

    for value in [False, True]:
        default_gen_sine.is_amplitude_autorange = value
        assert default_gen_sine.is_amplitude_autorange == value


def test_verify_amplitude(default_gen_sine):
    # Test type
    assert type(default_gen_sine.verify_amplitude(1)) is float
    # Test verify
    for ampl in [default_gen_sine.amplitude_min,
                 (default_gen_sine.amplitude_max-default_gen_sine.amplitude_min)/2,
                 default_gen_sine.amplitude_max]:
        assert default_gen_sine.verify_amplitude(ampl) == ampl
    # Test if amplitude clips
    assert default_gen_sine.verify_amplitude(default_gen_sine.amplitude_max+1.0) == default_gen_sine.amplitude_max


def test_is_offset_available(default_gen_sine):
    assert type(default_gen_sine.is_offset_available) is bool
    assert default_gen_sine.is_offset_available is True


def test_offset_min(default_gen_sine):
    assert type(default_gen_sine.offset_min) is float
    assert default_gen_sine.offset_min < 0


def test_offset_max(default_gen_sine):
    assert type(default_gen_sine.offset_max) is float
    assert default_gen_sine.offset_max > 0


def test_offset(default_gen_sine):
    # Test getter
    assert type(default_gen_sine.offset) is float

    # Test setter
    # Selectable offset depends on the signal amplitude
    low_limit = default_gen_sine.offset_min + default_gen_sine.amplitude
    high_limit = default_gen_sine.offset_max - default_gen_sine.amplitude
    for value in [low_limit, (high_limit-low_limit)/2, high_limit]:
        default_gen_sine.offset = value
        assert default_gen_sine.offset == pytest.approx(value)


def test_verify_offset(default_gen_sine):
    # Test type
    assert type(default_gen_sine.verify_offset(default_gen_sine.offset_max/2)) is float
    # Test verify
    low_limit = default_gen_sine.offset_min + default_gen_sine.amplitude
    high_limit = default_gen_sine.offset_max - default_gen_sine.amplitude
    for offset in [low_limit, (high_limit - low_limit) / 2, high_limit]:
        assert default_gen_sine.verify_offset(offset) == offset


def test_is_frequency_available(default_gen_sine):
    assert type(default_gen_sine.is_frequency_available) is bool
    assert default_gen_sine.is_frequency_available is True


def test_freq_min(default_gen_sine):
    assert type(default_gen_sine.freq_min) is float
    assert default_gen_sine.freq_min > 0
    assert default_gen_sine.freq_min < default_gen_sine.freq_max


def test_freq_max(default_gen_sine):
    assert type(default_gen_sine.freq_max) is float
    assert default_gen_sine.freq_max > 0
    assert default_gen_sine.freq_max > default_gen_sine.freq_min


def test_freq(default_gen_sine):
    # Test getter
    assert type(default_gen_sine.freq) is float
    assert default_gen_sine.freq >= default_gen_sine.freq_min
    assert default_gen_sine.freq <= default_gen_sine.freq_max

    # Test setter
    for freq in [default_gen_sine.freq_min,
                 (default_gen_sine.freq_max-default_gen_sine.freq_min)/2,
                 default_gen_sine.freq_max]:
        default_gen_sine.freq = freq
        assert default_gen_sine.freq == pytest.approx(freq)


def test_verify_frequency(default_gen_sine):
    # Test type
    assert type(default_gen_sine.verify_frequency(default_gen_sine.freq_max/2)) is float
    # Test verify
    assert math.isclose(default_gen_sine.verify_frequency(
        default_gen_sine.freq_max/2), default_gen_sine.freq_max/2, rel_tol=1e-6)
    # Test if the frequency clips
    assert default_gen_sine.verify_frequency(default_gen_sine.freq_max+1.0) == default_gen_sine.freq_max


def test_freq_modes_available(default_gen_sine):
    assert type(default_gen_sine.freq_modes_available) is tuple
    for freq_mode in default_gen_sine.freq_modes_available:
        assert freq_mode in default_gen_sine.FREQUENCY_MODES


def test_freq_mode(default_gen_sine):
    # Test getter
    assert default_gen_sine.freq_mode in default_gen_sine.FREQUENCY_MODES

    # Test setter
    for mode in default_gen_sine.freq_modes_available:
        default_gen_sine.freq_mode = mode
        assert default_gen_sine.freq_mode == mode


def test_is_phase_available(default_gen_sine):
    assert type(default_gen_sine.is_phase_available) is bool
    assert default_gen_sine.is_phase_available is True


def test_phase_min(default_gen_sine):
    assert type(default_gen_sine.phase_min) is float
    assert default_gen_sine.phase_min >= 0
    assert default_gen_sine.phase_min <= 360


def test_phase_max(default_gen_sine):
    assert type(default_gen_sine.phase_max) is float
    assert default_gen_sine.phase_max >= 0
    assert default_gen_sine.phase_max <= 360


def test_phase(default_gen_sine):
    # Test getter
    assert type(default_gen_sine.phase) is float
    assert default_gen_sine.phase >= default_gen_sine.phase_min
    assert default_gen_sine.phase <= default_gen_sine.phase_max

    # Test setter
    for phase in [default_gen_sine.phase_min,
                  (default_gen_sine.phase_max-default_gen_sine.phase_min)/2,
                  default_gen_sine.phase_max]:
        default_gen_sine.phase = phase
        assert default_gen_sine.phase == phase


def test_verify_phase(default_gen_sine):
    # Test type
    assert type(default_gen_sine.verify_phase(default_gen_sine.phase_max/2)) is float
    # Test verify
    assert math.isclose(
        default_gen_sine.verify_phase(default_gen_sine.phase_max / 2),
        default_gen_sine.phase_max / 2,
        rel_tol=1e-6)
    # Test if the phase overruns
    assert default_gen_sine.verify_phase(
        default_gen_sine.phase_max + 1.0) == default_gen_sine.phase_max


def test_is_symmetry_available(default_gen_sine):
    assert type(default_gen_sine.is_symmetry_available) is bool
    assert default_gen_sine.is_symmetry_available is True


def test_symmetry_min(default_gen_sine):
    assert type(default_gen_sine.symmetry_min) is float
    assert default_gen_sine.symmetry_min >= 0
    assert default_gen_sine.symmetry_min <= 1


def test_symmetry_max(default_gen_sine):
    assert type(default_gen_sine.symmetry_max) is float
    assert default_gen_sine.symmetry_max >= 0
    assert default_gen_sine.symmetry_min <= 1
    assert default_gen_sine.symmetry_max > default_gen_sine.symmetry_min


def test_symmetry(default_gen_sine):
    # Test getter
    assert type(default_gen_sine.symmetry) is float
    assert default_gen_sine.symmetry >= 0
    assert default_gen_sine.symmetry <= 1

    # Test setter
    values = [0.0, 0.25, 0.5, 0.75, 1.0]
    for value in values:
        default_gen_sine.symmetry = value
        assert default_gen_sine.symmetry == value


def test_verify_symmetry(default_gen_sine):
    # Test type
    assert type(default_gen_sine.verify_symmetry(default_gen_sine.symmetry_max/2)) is float
    # Test verify
    assert math.isclose(default_gen_sine.verify_symmetry(
        default_gen_sine.symmetry_max / 2),
        default_gen_sine.symmetry_max / 2,
        rel_tol=1e-6)
    # Test if the symmetry clips
    assert default_gen_sine.verify_symmetry(
        default_gen_sine.symmetry_max + 1.0) == default_gen_sine.symmetry_max


def test_is_pulse_width_available(default_gen_pulse):
    assert type(default_gen_pulse.is_pulse_width_available) is bool
    assert default_gen_pulse.is_pulse_width_available is True


def test_pulse_width_min(default_gen_pulse):
    assert type(default_gen_pulse.pulse_width_min) is float
    assert default_gen_pulse.pulse_width_min >= 0


def test_pulse_width_max(default_gen_pulse):
    assert type(default_gen_pulse.pulse_width_max) is float
    assert default_gen_pulse.pulse_width_max >= 0
    assert default_gen_pulse.pulse_width_max >= default_gen_pulse.pulse_width_min


def test_pulse_width(default_gen_pulse):
    # Test getter
    assert type(default_gen_pulse.pulse_width) is float
    assert default_gen_pulse.pulse_width >= 0
    assert default_gen_pulse.pulse_width >= default_gen_pulse.pulse_width_min
    assert default_gen_pulse.pulse_width <= default_gen_pulse.pulse_width_max

    # Test setter
    values = [default_gen_pulse.pulse_width_min,
              (default_gen_pulse.pulse_width_max-default_gen_pulse.pulse_width_min)/2,
              default_gen_pulse.pulse_width_max]
    for value in values:
        default_gen_pulse.pulse_width = value
        assert default_gen_pulse.pulse_width == pytest.approx(value)


def test_verify_pulse_width(default_gen_pulse):
    # Test type
    assert type(default_gen_pulse.verify_pulse_width(default_gen_pulse.pulse_width_max/2)) is float
    # Test verify
    assert math.isclose(default_gen_pulse.verify_pulse_width(
        default_gen_pulse.pulse_width_max / 2),
        default_gen_pulse.pulse_width_max / 2,
        rel_tol=1e-6)
    # Test if the pulse width clips
    assert default_gen_pulse.verify_pulse_width(
        default_gen_pulse.pulse_width_max + 1.0) == default_gen_pulse.pulse_width_max


def test_is_arb_data_available(default_gen_arb):
    assert type(default_gen_arb.is_arb_data_available) is bool
    assert default_gen_arb.is_arb_data_available is True


def test_arb_data_length_min(default_gen_arb):
    assert type(default_gen_arb.arb_data_length_min) is int
    assert default_gen_arb.arb_data_length_min > 0


def test_arb_data_length_max(default_gen_arb):
    assert type(default_gen_arb.arb_data_length_max) is int
    assert default_gen_arb.arb_data_length_max > default_gen_arb.arb_data_length_min


def test_arb_data_length(default_gen_arb):
    assert type(default_gen_arb.arb_data_length) is int
    assert default_gen_arb.arb_data_length >= default_gen_arb.arb_data_length_min
    assert default_gen_arb.arb_data_length <= default_gen_arb.arb_data_length_max


def test_verify_arb_data_length(default_gen_arb):
    # Test type
    assert type(default_gen_arb.verify_arb_data_length(default_gen_arb.arb_data_length_max // 2)) is int
    # Test verify
    assert default_gen_arb.verify_arb_data_length(
        default_gen_arb.arb_data_length_max // 2) == default_gen_arb.arb_data_length_max // 2


def test_arb_data(default_gen_arb):
    # Test with empty list
    default_gen_arb.arb_data([])
    assert default_gen_arb.arb_data_length == 1

    # Test with data
    default_gen_arb.arb_data([0.0, 1.0, 2.0, 3.0, 4.0])
    assert default_gen_arb.arb_data_length == 5


def test_arb_data_raw_type(default_gen_arb):
    assert type(default_gen_arb.arb_raw_type) is str
    assert default_gen_arb.arb_raw_type in default_gen_arb.RAW_DATA_TYPES.keys()


def test_arb_data_raw_min(default_gen_arb):
    assert type(default_gen_arb.arb_data_raw_min) is int
    assert default_gen_arb.arb_data_raw_min >= 0


def test_arb_data_raw_max(default_gen_arb):
    assert type(default_gen_arb.arb_data_raw_max) is int
    assert default_gen_arb.arb_data_raw_max >= 0
    assert default_gen_arb.arb_data_raw_max >= default_gen_arb.arb_data_raw_min


def test_arb_data_raw_max(default_gen_arb):
    assert type(default_gen_arb.arb_data_raw_max) is int
    assert default_gen_arb.arb_data_raw_max >= 0
    assert default_gen_arb.arb_data_raw_max >= default_gen_arb.arb_data_raw_min


def test_arb_data_raw_zero(default_gen_arb):
    assert type(default_gen_arb.arb_data_raw_zero) is int
    assert default_gen_arb.arb_data_raw_zero >= default_gen_arb.arb_data_raw_min
    assert default_gen_arb.arb_data_raw_zero <= default_gen_arb.arb_data_raw_max


def test_arb_data_raw_range(default_gen_arb):
    assert type(default_gen_arb.arb_data_raw_range) is tuple
    assert default_gen_arb.arb_data_raw_range[0] == default_gen_arb.arb_data_raw_min
    assert default_gen_arb.arb_data_raw_range[1] == default_gen_arb.arb_data_raw_zero
    assert default_gen_arb.arb_data_raw_range[2] == default_gen_arb.arb_data_raw_max


def test_modes_native_available(default_gen_sine):
    assert type(default_gen_sine.modes_native_available) is tuple
    for mode in default_gen_sine.modes_native_available:
        assert mode in default_gen_sine.GENERATOR_MODES


def test_modes_available(default_gen_sine):
    assert type(default_gen_sine.modes_available) is tuple
    for mode in default_gen_sine.modes_available:
        assert mode in default_gen_sine.GENERATOR_MODES


def test_mode(default_gen_sine):
    # Test getter
    assert default_gen_sine.mode in default_gen_sine.GENERATOR_MODES

    # Test setter
    for mode in default_gen_sine.modes_available:
        default_gen_sine.mode = mode
        assert default_gen_sine.mode == mode


def test_is_burst_active(default_gen_burst):
    assert type(default_gen_burst.is_burst_active) is bool
    # Maybe test after starting burst?


def test_burst_cnt_min(default_gen_burst):
    assert type(default_gen_burst.burst_cnt_min) is int
    assert default_gen_burst.burst_cnt_min >= 0


def test_burst_cnt_max(default_gen_burst):
    assert type(default_gen_burst.burst_cnt_max) is int
    assert default_gen_burst.burst_cnt_max > default_gen_burst.burst_cnt_min


def test_burst_cnt(default_gen_burst):
    # Test getter
    assert type(default_gen_burst.burst_cnt) is int
    assert default_gen_burst.burst_cnt >= default_gen_burst.burst_cnt_min
    assert default_gen_burst.burst_cnt <= default_gen_burst.burst_cnt_max

    # Test setter
    for value in [default_gen_burst.burst_cnt_min,
                  math.ceil((default_gen_burst.burst_cnt_max - default_gen_burst.burst_cnt_min)/2),
                  default_gen_burst.burst_cnt_max]:
        default_gen_burst.burst_cnt = value
        assert default_gen_burst.burst_cnt == value


def test_burst_sample_cnt_min(default_gen_burst_sample):
    assert type(default_gen_burst_sample.burst_sample_cnt_min) is int
    assert default_gen_burst_sample.burst_sample_cnt_min >= 0


def test_burst_sample_cnt_max(default_gen_burst_sample):
    assert type(default_gen_burst_sample.burst_sample_cnt_max) is int
    assert default_gen_burst_sample.burst_sample_cnt_max > default_gen_burst_sample.burst_sample_cnt_min


def test_burst_sample_cnt(default_gen_burst_sample):
    # Burst sample count is initially always 0 even though it is not
    # in the range. Setting it to some value sets or clips it accordingly
    default_gen_burst_sample.burst_sample_cnt = 2
    # Test getter
    assert type(default_gen_burst_sample.burst_sample_cnt) is int
    assert default_gen_burst_sample.burst_sample_cnt >= default_gen_burst_sample.burst_sample_cnt_min
    assert default_gen_burst_sample.burst_sample_cnt <= default_gen_burst_sample.burst_sample_cnt_max

    # Test setter
    for value in [default_gen_burst_sample.burst_sample_cnt_min,
                  math.ceil((default_gen_burst_sample.burst_sample_cnt_max -
                             default_gen_burst_sample.burst_sample_cnt_min) / 2),
                  default_gen_burst_sample.burst_sample_cnt_max]:
        default_gen_burst_sample.burst_sample_cnt = value
        assert default_gen_burst_sample.burst_sample_cnt == value


def test_burst_segment_cnt_min(default_gen_burst_segment):
    assert type(default_gen_burst_segment.burst_segment_cnt_min) is int
    assert default_gen_burst_segment.burst_segment_cnt_min >= 0


def test_burst_segment_cnt_max(default_gen_burst_segment):
    assert type(default_gen_burst_segment.burst_segment_cnt_max) is int
    assert default_gen_burst_segment.burst_segment_cnt_max > default_gen_burst_segment.burst_segment_cnt_min


def test_burst_segment_cnt(default_gen_burst_segment):
    # Test getter
    assert type(default_gen_burst_segment.burst_segment_cnt) is int
    # Note: burst_segment_cnt can be read as 0, even if burst_segment_cnt_min is 1
    assert default_gen_burst_segment.burst_segment_cnt >= 0
    assert default_gen_burst_segment.burst_segment_cnt <= default_gen_burst_segment.burst_segment_cnt_max

    # Test setter
    for value in [default_gen_burst_segment.burst_segment_cnt_min,
                  math.ceil((default_gen_burst_segment.burst_segment_cnt_max -
                         default_gen_burst_segment.burst_segment_cnt_min) / 2),
                  default_gen_burst_segment.burst_segment_cnt_max]:
        default_gen_burst_segment.burst_segment_cnt = value
        assert default_gen_burst_segment.burst_segment_cnt == value


def test_verify_burst_segment_cnt(default_gen_burst_segment):
    # Test type
    assert type(default_gen_burst_segment.verify_burst_segment_cnt(default_gen_burst_segment.burst_segment_cnt_max//2)) is int
    # Test verify
    assert default_gen_burst_segment.verify_burst_segment_cnt(
        default_gen_burst_segment.burst_segment_cnt_max // 2) == default_gen_burst_segment.burst_segment_cnt_max // 2
    # Test if the burst segment count clips
    assert default_gen_burst_segment.verify_burst_segment_cnt(
        default_gen_burst_segment.burst_segment_cnt_max + 1) == default_gen_burst_segment.burst_segment_cnt_max



import pytest
from handyscope.triggerOutput import TriggerOutput


def test_is_enabled(device):
    # test getter
    for trig_out in device.trig_outs:
        assert type(trig_out.is_enabled) is bool

    # test setter by changing the value and reading it back
    for trig_out in device.trig_outs:
        trig_out.is_enabled = True
        assert trig_out.is_enabled is True
        trig_out.is_enabled = False
        assert trig_out.is_enabled is False


def test_trigger_id(device):
    for trig_out in device.trig_outs:
        assert trig_out.trigger_id in TriggerOutput.TRIGGER_IDS


def test_name(device):
    for trig_out in device.trig_outs:
        # Names correspond to keys in TRIGGER_IDS
        assert trig_out.name in TriggerOutput.TRIGGER_IDS


def test_events_available(device):
    for trig_out in device.trig_outs:
        for event in trig_out.events_available:
            assert event in TriggerOutput.TRIGGER_EVENTS


def test_event(device):
    for trig_out in device.trig_outs:
        # test getter
        assert trig_out.event in TriggerOutput.TRIGGER_EVENTS

        # test setter by changing a value and reading it back
        # possible values are accessible via property events
        for event in trig_out.events_available:
            trig_out.event = event
            assert trig_out.event is event


def test_force_trig(device):
    for trig_out in device.trig_outs:
        trig_out.is_enabled = True
        # HS5 doesn't support forced output triggers
        with pytest.raises(OSError) as err:
            trig_out.force_trig()
        assert err.value.args[0] == "[-2]: NOT_SUPPORTED"



import ctypes
import time
import warnings

import numpy as np

from handyscope.device import Device
from handyscope.library import libtiepie
from handyscope.oscilloscopeChannel import OscilloscopeChannel


class Oscilloscope(Device):
    """Class for an oscilloscope.

    Attributes:
        MEASURE_MODES (dict): dict which maps measure modes as strs to their
                              libtiepie int version
        AUTO_RESOLUTIONS (dict): dict which maps auto resolutions as strs to
                                 their libtiepie int version
        CLOCK_SOURCES (dict): dict which maps clock sources as strs to their
                              libtiepie int version
        CLOCK_OUTPUTS (dict): dict which maps clock outputs as strs to their
                              libtiepie int version
        CONNECTION_STATES (dict): dict which maps connection states as strs to
                                  their libtiepie int version
    """
    MEASURE_MODES = {"unknown": 0,
                     "stream": 1,
                     "block": 2}

    AUTO_RESOLUTIONS = {"unknown": 0,
                        "disabled": 1,
                        "native only": 2,
                        "all": 4}

    CLOCK_SOURCES = {"unknown": 0,
                     "external": 1,
                     "internal": 2}

    CLOCK_OUTPUTS = {"unknown": 0,
                     "disabled": 1,
                     "sample": 2,
                     "fixed": 4}

    CONNECTION_STATES = {"undefined": 0,
                         "connected": 1,
                         "disconnected": 2}

    DATA_TYPES = {"int8": ctypes.c_int8,
                  "int16": ctypes.c_int16,
                  "int32": ctypes.c_int32,
                  "int64": ctypes.c_int64,
                  "uint8": ctypes.c_uint8,
                  "uint16": ctypes.c_uint16,
                  "uint32": ctypes.c_uint32,
                  "uint64": ctypes.c_uint64,
                  "float32": ctypes.c_float,
                  "float64": ctypes.c_double}

    TRIG_HOLDOFF_ALL_PRE_SAMPLES = 0xffffffffffffffff

    _device_type = "Osc"

    def __init__(self, instr_id, id_kind="product id"):
        """Constructor for an oscilloscope.

        Args:
            instr_id (int or str): Device list index, product ID (listed in dict
                                   PRODUCT_IDS) or serial number
            id_kind (str): the kind of the given instr_id
                          (listed in dict ID_KINDS)
        """
        super().__init__(instr_id, id_kind, self._device_type)

        # Initialize channels
        self._channels = tuple(
            OscilloscopeChannel(self._dev_handle, ch_idx) for ch_idx in
            range(self.channel_cnt))

    @property
    def channel_cnt(self):
        """Get the channel count.

        Returns:
            int: The channel count
        """
        return libtiepie.ScpGetChannelCount(self._dev_handle)

    @property
    def channels(self):
        """Tuple of all channels.

        Returns:
            tuple: Tuple of oscilloscope channels 
                   (:py:class:`handyscope.oscilloscopeChannel.OscilloscopeChannel`)
        """
        return self._channels

    def _get_sample_cnts(self):
        """Get information on the sample counts.

        The recorded samples are divided in pre and post samples. If multiple
        triggers occur in a short time, not all pre samples might be valid.
        This method calculates the start count and the total number of valid
        samples.

        Returns:
            tuple: sample start count and valid sample count as ints
        """
        # Calc number of valid samples
        if self.measure_mode == 'block':
            post_sample_cnt = round(
                (1.0 - self.pre_sample_ratio) * self.record_length)
            valid_sample_cnt = post_sample_cnt + self.valid_pre_sample_cnt
            # Calc sample start count
            sample_start_cnt = self.record_length - valid_sample_cnt
        else:
            sample_start_cnt = 0
            valid_sample_cnt = self.record_length

        return sample_start_cnt, valid_sample_cnt

    def retrieve(self, channel_nos=None, raw=False):
        """Retrieve measured samples.

        Previously to retrieving data, a measurement has to be started.
        If no channel numbers are given, all enabled channels are retrieved.

        Args:
            channel_nos (list): (optional) iterable with channel numbers to
                                retrieve, or None
            raw (bool): True, if raw data should be returned.

        Returns:
            list: List with entries for each channel. An entry contains None,
                  if the channel is disabled, otherwise a list of samples.
        """
        # If no channel numbers are given, get the active ones
        if channel_nos is None:
            channel_nos = []
            for channel in self.channels:
                if channel.is_enabled:
                    channel_nos.append(channel._idx + 1)
        # Else check that the given channels are enabled
        else:
            for channel_no in channel_nos:
                if self.channels[channel_no - 1].is_enabled is False:
                    raise ValueError(
                        "The given channel %d is not enabled. It has to be enabled before " % channel_no
                        + "oscilloscope.start() is called to get valid values!")

        # Check that there is at least one entry in channel_nos
        if not channel_nos:
            raise ValueError(
                "No channel is enabled for measurement or the supplied channel number list is empty.")

        # Get number of valid samples
        sample_start_cnt, valid_sample_cnt = self._get_sample_cnts()

        # Initialize buffer
        channel_cnt = max(channel_nos)
        buffers = [None] * channel_cnt
        pointer_array = libtiepie.HlpPointerArrayNew(channel_cnt)
        for idx in range(channel_cnt):
            if idx + 1 in channel_nos:
                if raw:
                    raw_type = self.channels[idx].raw_data_type
                    c_type = self.DATA_TYPES[raw_type]
                else:
                    c_type = ctypes.c_float
                buffers[idx] = (c_type * valid_sample_cnt)()
                libtiepie.HlpPointerArraySet(pointer_array, idx,
                                             ctypes.byref(buffers[idx]))

        if raw:
            libtiepie.ScpGetDataRaw(self._dev_handle, pointer_array,
                                    channel_cnt, sample_start_cnt,
                                    valid_sample_cnt)
        else:
            libtiepie.ScpGetData(self._dev_handle, pointer_array, channel_cnt,
                                 sample_start_cnt, valid_sample_cnt)

        # Free pointer array
        libtiepie.HlpPointerArrayDelete(pointer_array)

        # Cast ctypes float array to normal python lists
        data = [None if channel_data is None else list(channel_data) for
                channel_data in buffers]

        return data

    def retrieve_ch1(self):
        """Retrieve measured samples of channel 1.

        Previously to retrieving data, a measurement has to be started.

        Returns:
            list: List with an entry for channel 1. The entry contains None,
                  if channel 1 is disabled, otherwise a list of samples.
        """
        # Check availability
        if self.channels[0].is_enabled:
            # Get number of valid samples
            sample_start_cnt, valid_sample_cnt = self._get_sample_cnts()

            # Init buffer
            buffer = (ctypes.c_float * valid_sample_cnt)()

            libtiepie.ScpGetData1Ch(self._dev_handle, buffer, sample_start_cnt,
                                    valid_sample_cnt)

            # Cast ctypes float array to normal python list
            data = list(buffer)

            return [data]
        else:
            return [None]

    def retrieve_ch1_to_ch2(self):
        """Retrieve measured samples of channel 1 and 2.

        Previously to retrieving data, a measurement has to be started.
        
        Returns:
            list: List with entries for each channel. An entry contains None,
                  if the channel is disabled, otherwise a list of samples.
        """
        # Get number of valid samples
        sample_start_cnt, valid_sample_cnt = self._get_sample_cnts()

        # Init buffers list for channel 1 & 2
        buffers = [None] * 2

        for idx in range(len(buffers)):
            # Check availability
            if idx < len(self.channels):
                if self.channels[idx].is_enabled:
                    buffers[idx] = (ctypes.c_float * valid_sample_cnt)()

        libtiepie.ScpGetData2Ch(self._dev_handle, *buffers, sample_start_cnt,
                                valid_sample_cnt)

        # Cast ctypes float array to normal python lists
        data = [None if channel_data is None else list(channel_data) for
                channel_data in buffers]

        return data

    def retrieve_ch1_to_ch3(self):
        """Retrieve measured samples of channel 1 to 3.

        Previously to retrieving data, a measurement has to be started.

        Returns:
            list: List with entries for each channel. An entry contains None,
                  if the channel is disabled, otherwise a list of samples.
        """
        # Get number of valid samples
        sample_start_cnt, valid_sample_cnt = self._get_sample_cnts()

        # Init buffers list for channel 1 to 3
        buffers = [None] * 3

        for idx in range(len(buffers)):
            # Check availability
            if idx < len(self.channels):
                if self.channels[idx].is_enabled:
                    buffers[idx] = (ctypes.c_float * valid_sample_cnt)()

        libtiepie.ScpGetData3Ch(self._dev_handle, *buffers, sample_start_cnt,
                                valid_sample_cnt)

        # Cast ctypes float array to normal python lists
        data = [None if channel_data is None else list(channel_data) for
                channel_data in buffers]

        return data

    def retrieve_ch1_to_ch4(self):
        """Retrieve measured samples of channel 1 to 4.

        Previously to retrieving data, a measurement has to be started.
        
        Returns:
            list: List with entries for each channel. An entry contains None,
                  if the channel is disabled, otherwise a list of samples.
        """
        # Get number of valid samples
        sample_start_cnt, valid_sample_cnt = self._get_sample_cnts()

        # Init buffers list for channel 1 to 4
        buffers = [None] * 4

        for idx in range(len(buffers)):
            # Check availability
            if idx < len(self.channels):
                if self.channels[idx].is_enabled:
                    buffers[idx] = (ctypes.c_float * valid_sample_cnt)()

        libtiepie.ScpGetData4Ch(self._dev_handle, *buffers, sample_start_cnt,
                                valid_sample_cnt)

        # Cast ctypes float array to normal python lists
        data = [None if channel_data is None else list(channel_data) for
                channel_data in buffers]

        return data

    def retrieve_ch1_to_ch5(self):
        """Retrieve measured samples of channel 1 to 5.

        Previously to retrieving data, a measurement has to be started.
        
        Not tested.

        Returns:
            list: List with entries for each channel. An entry contains None,
                  if the channel is disabled, otherwise a list of samples.
        """
        # Get number of valid samples
        sample_start_cnt, valid_sample_cnt = self._get_sample_cnts()

        # Init buffers list for channel 1 to 5
        buffers = [None] * 5

        for idx in range(len(buffers)):
            # Check availability
            if idx < len(self.channels):
                if self.channels[idx].is_enabled:
                    buffers[idx] = (ctypes.c_float * valid_sample_cnt)()

        libtiepie.ScpGetData5Ch(self._dev_handle, *buffers, sample_start_cnt,
                                valid_sample_cnt)

        # Cast ctypes float array to normal python lists
        data = [None if channel_data is None else list(channel_data) for
                channel_data in buffers]

        return data

    def retrieve_ch1_to_ch6(self):
        """Retrieve measured samples of channel 1 to 6.

        Previously to retrieving data, a measurement has to be started.
        
        Not tested.

        Returns:
            list: List with entries for each channel. An entry contains None,
                  if the channel is disabled, otherwise a list of samples.
        """
        # Get number of valid samples
        sample_start_cnt, valid_sample_cnt = self._get_sample_cnts()

        # Init buffers list for channel 1 to 6
        buffers = [None] * 6

        for idx in range(len(buffers)):
            # Check availability
            if idx < len(self.channels):
                if self.channels[idx].is_enabled:
                    buffers[idx] = (ctypes.c_float * valid_sample_cnt)()

        libtiepie.ScpGetData6Ch(self._dev_handle, *buffers, sample_start_cnt,
                                valid_sample_cnt)

        # Cast ctypes float array to normal python lists
        data = [None if channel_data is None else list(channel_data) for
                channel_data in buffers]

        return data

    def retrieve_ch1_to_ch7(self):
        """Retrieve measured samples of channel 1 to 7.

        Previously to retrieving data, a measurement has to be started.
        
        Not tested.

        Returns:
            list: List with entries for each channel. An entry contains None, 
                  if the channel is disabled, otherwise a list of samples.
        """
        # Get number of valid samples
        sample_start_cnt, valid_sample_cnt = self._get_sample_cnts()

        # Init buffers list for channel 1 to 7
        buffers = [None] * 7

        for idx in range(len(buffers)):
            # Check availability
            if idx < len(self.channels):
                if self.channels[idx].is_enabled:
                    buffers[idx] = (ctypes.c_float * valid_sample_cnt)()

        libtiepie.ScpGetData7Ch(self._dev_handle, *buffers, sample_start_cnt,
                                valid_sample_cnt)

        # Cast ctypes float array to normal python lists
        data = [None if channel_data is None else list(channel_data) for
                channel_data in buffers]

        return data

    def retrieve_ch1_to_ch8(self):
        """Retrieve measured samples of channel 1 to 8.

        Previously to retrieving data, a measurement has to be started.
        
        Not tested.

        Returns:
            list: List with entries for each channel. An entry contains None, 
                  if the channel is disabled, otherwise a list of samples.
        """
        # Get number of valid samples
        sample_start_cnt, valid_sample_cnt = self._get_sample_cnts()

        # Init buffers list for channel 1 to 8
        buffers = [None] * 8

        for idx in range(len(buffers)):
            # Check availability
            if idx < len(self.channels):
                if self.channels[idx].is_enabled:
                    buffers[idx] = (ctypes.c_float * valid_sample_cnt)()

        libtiepie.ScpGetData8Ch(self._dev_handle, *buffers, sample_start_cnt,
                                valid_sample_cnt)

        # Cast ctypes float array to normal python lists
        data = [None if channel_data is None else list(channel_data) for
                channel_data in buffers]

        return data

    @property
    def valid_pre_sample_cnt(self):
        """Get the count of valid pre samples.

        Returns:
            int: Count of valid pre samples.
        """
        return libtiepie.ScpGetValidPreSampleCount(self._dev_handle)

    def start(self):
        """Start a measurement.

        Returns:
            bool: True if successful, False otherwise.
        """
        return libtiepie.ScpStart(self._dev_handle) == 1

    def stop(self):
        """Stop a measurement.

        Returns:
            bool: True if successful, False otherwise.
        """
        return libtiepie.ScpStop(self._dev_handle) == 1

    def force_trig(self):
        """Force a trigger.

        Returns:
            bool: True if successful, False otherwise.
        """
        return libtiepie.ScpForceTrigger(self._dev_handle) == 1

    @property
    def measure_modes_available(self):
        """Get the available measure modes.

        Returns:
            tuple: Available measure modes (keys of :py:attr:`handyscope.oscilloscope.Oscilloscope.MEASURE_MODES`)
        """
        raw_modes = libtiepie.ScpGetMeasureModes(self._dev_handle)
        _modes = []

        # If no measure modes are available, return unknown
        if raw_modes == self.MEASURE_MODES["unknown"]:
            _modes.append("unknown")
        # Else do a detailed analysis...
        else:
            # ... by iterating over every possible kind...
            for key in self.MEASURE_MODES:
                # ... and ignoring "unknown" (already handled above)
                if key == "unknown":
                    pass
                elif raw_modes & self.MEASURE_MODES[key] == self.MEASURE_MODES[
                        key]:
                    _modes.append(key)

        return tuple(_modes)

    @property
    def measure_mode(self):
        """Get or set the current measure mode (keys of
        :py:attr:`handyscope.oscilloscope.Oscilloscope.MEASURE_MODES`)"""
        mode_int = libtiepie.ScpGetMeasureMode(self._dev_handle)
        for key in self.MEASURE_MODES:
            if mode_int == self.MEASURE_MODES[key]:
                return key

        raise ValueError("Unknown measure mode: %d" % mode_int)

    @measure_mode.setter
    def measure_mode(self, value):
        libtiepie.ScpSetMeasureMode(
            self._dev_handle, self.MEASURE_MODES[value])

    @property
    def is_running(self):
        """Check if the oscilloscope is running.

        Returns:
            bool: Truef if oscilloscope is running, False otherwise.
        """
        return libtiepie.ScpIsRunning(self._dev_handle) == 1

    @property
    def is_triggered(self):
        """Check if the oscilloscope is triggered.

        Returns:
            bool: True if oscilloscope is triggered, False otherwise.
        """
        return libtiepie.ScpIsTriggered(self._dev_handle) == 1

    @property
    def is_timeout_trig(self):
        """Check if the oscilloscope is triggered by a timeout.

        Returns:
            bool: True if oscilloscope is triggered by a timeout,
                  False otherwise.
        """
        return libtiepie.ScpIsTimeOutTriggered(self._dev_handle) == 1

    @property
    def is_force_trig(self):
        """Check if the trigger of the oscilloscope is forced.

        Returns:
            bool: True if trigger is forced, False otherwise.
        """
        return libtiepie.ScpIsForceTriggered(self._dev_handle) == 1

    @property
    def is_data_ready(self):
        """Check if data is ready.

        Returns:
            bool: True if data ready, False otherwise.
        """
        return libtiepie.ScpIsDataReady(self._dev_handle) == 1

    @property
    def is_data_overflow(self):
        """Check if there is a data overflow.

        Returns:
            bool: True if data overflow, False otherwise.
        """
        return libtiepie.ScpIsDataOverflow(self._dev_handle) == 1

    @property
    def resolutions_available(self):
        """Get available ADC resolutions.

        Returns:
            tuple: Available ADC resolutions in bits.
        """
        # get length of list
        res_len = libtiepie.ScpGetResolutions(self._dev_handle, None, 0)

        # initialize uint8 array
        res = (ctypes.c_uint8 * res_len)()

        # write the actual data to the array
        libtiepie.ScpGetResolutions(self._dev_handle, ctypes.byref(res),
                                    res_len)

        # convert to a normal python list
        res = list(res)

        return tuple(res)

    @property
    def resolution(self):
        """Get or set the current ADC resolution in bit."""
        return libtiepie.ScpGetResolution(self._dev_handle)

    @resolution.setter
    def resolution(self, value):
        libtiepie.ScpSetResolution(self._dev_handle, value)

    @property
    def is_resolution_enhanced(self):
        """Check if the ADC resolution is enhanced.

        Returns:
            bool: True if ADC resolution is enhanced, False otherwise.
        """
        return libtiepie.ScpIsResolutionEnhanced(self._dev_handle) == 1

    @property
    def auto_resolutions_available(self):
        """Get available auto resolutions.

        Returns:
            tuple: Available auto resolutions (keys of :py:attr:`handyscope.oscilloscope.Oscilloscope.AUTO_RESOLUTIONS`)
        """
        raw_res = libtiepie.ScpGetAutoResolutionModes(self._dev_handle)
        _res = []

        # If no auto resolution modes are available, return unknown
        if raw_res == self.AUTO_RESOLUTIONS["unknown"]:
            _res.append("unknown")
        # Else do a detailed analysis...
        else:
            # ... by iterating over every possible kind...
            for key in self.AUTO_RESOLUTIONS:
                # ... and ignoring "unknown" (already handled above)
                if key == "unknown":
                    pass
                elif raw_res & self.AUTO_RESOLUTIONS[key] == \
                        self.AUTO_RESOLUTIONS[key]:
                    _res.append(key)

        return tuple(_res)

    @property
    def auto_resolution(self):
        """Get or set the current auto resolution mode (key of
        :py:attr:`handyscope.oscilloscope.Oscilloscope.AUTO_RESOLUTIONS`)
        """
        raw_res = libtiepie.ScpGetAutoResolutionMode(self._dev_handle)
        for key, value in self.AUTO_RESOLUTIONS.items():
            if raw_res == value:
                return key

        raise ValueError("Unknown auto resolution mode: %d" % raw_res)

    @auto_resolution.setter
    def auto_resolution(self, value):
        libtiepie.ScpSetAutoResolutionMode(self._dev_handle,
                                           self.AUTO_RESOLUTIONS[value])

    @property
    def clock_sources_available(self):
        """Get available clock sources.

        Returns:
            tuple: Available clock sources (keys of :py:attr:`handyscope.oscilloscope.Oscilloscope.CLOCK_SOURCES`)
        """
        raw_srcs = libtiepie.ScpGetClockSources(self._dev_handle)
        srcs = []

        if raw_srcs == self.CLOCK_SOURCES["unknown"]:
            srcs.append("unknown")
        else:
            for key, value in self.CLOCK_SOURCES.items():
                if key == "unknown":
                    continue
                else:
                    if raw_srcs & value == value:
                        srcs.append(key)

        return tuple(srcs)

    @property
    def clock_source(self):
        """Get or set the current clock source (key of
        :py:attr:`handyscope.oscilloscope.Oscilloscope.CLOCK_SOURCES`)"""
        src = libtiepie.ScpGetClockSource(self._dev_handle)
        for key, value in self.CLOCK_SOURCES.items():
            if src == value:
                return key

        raise ValueError("Unknown clock source: %d" % src)

    @clock_source.setter
    def clock_source(self, value):
        libtiepie.ScpSetClockSource(
            self._dev_handle, self.CLOCK_SOURCES[value])

    @property
    def clock_outputs_available(self):
        """Get available clock outputs.

        Returns:
            tuple: Available clock outputs (keys of :py:attr:`handyscope.oscilloscope.Oscilloscope.CLOCK_OUTPUTS`)
        """
        raw_outs = libtiepie.ScpGetClockOutputs(self._dev_handle)
        outs = []

        if raw_outs == self.CLOCK_OUTPUTS["unknown"]:
            outs.append("unknown")
        else:
            for key, value in self.CLOCK_OUTPUTS.items():
                if key == "unknown":
                    pass
                else:
                    if raw_outs & value == value:
                        outs.append(key)

        return tuple(outs)

    @property
    def clock_output(self):
        """Get or set the current clock output (key of
        :py:attr:`handyscope.oscilloscope.Oscilloscope.CLOCK_OUTPUTS`)"""
        out = libtiepie.ScpGetClockOutput(self._dev_handle)
        for key, value in self.CLOCK_OUTPUTS.items():
            if out == value:
                return key

        raise ValueError("Unknown clock output: %d" % out)

    @clock_output.setter
    def clock_output(self, value):
        libtiepie.ScpSetClockOutput(
            self._dev_handle, self.CLOCK_OUTPUTS[value])

    @property
    def clock_source_frequencies_available(self):
        """Get the available frequencies for the clock source."""
        frequencies_len = libtiepie.ScpGetClockSourceFrequencies(
            self._dev_handle,
            None,
            0)

        # Initialize double array
        frequencies = (ctypes.c_double * frequencies_len)()

        # Write the actual data to the array
        libtiepie.ScpGetClockSourceFrequencies(self._dev_handle,
                                               ctypes.byref(frequencies),
                                               frequencies_len)

        # Convert to a normal python list
        frequencies = tuple(frequencies)

        return frequencies

    @property
    def clock_source_frequency(self):
        """Get or set the clock source frequency in Hz."""
        return libtiepie.ScpGetClockSourceFrequency(self._dev_handle)

    @clock_source_frequency.setter
    def clock_source_frequency(self, value):
        libtiepie.ScpSetClockSourceFrequency(self._dev_handle, value)

    @property
    def clock_output_frequencies_available(self):
        """Get the available frequencies for the clock output."""
        frequencies_len = libtiepie.ScpGetClockOutputFrequencies(
            self._dev_handle,
            None,
            0)

        # Initialize double array
        frequencies = (ctypes.c_double * frequencies_len)()

        # Write the actual data to the array
        libtiepie.ScpGetClockOutputFrequencies(self._dev_handle,
                                               ctypes.byref(frequencies),
                                               frequencies_len)

        # Convert to a normal python list
        frequencies = tuple(frequencies)

        return frequencies

    @property
    def clock_output_frequency(self):
        """Get or set the clock source frequency in Hz."""
        return libtiepie.ScpGetClockOutputFrequency(self._dev_handle)

    @clock_output_frequency.setter
    def clock_output_frequency(self, value):
        libtiepie.ScpSetClockOutputFrequency(self._dev_handle, value)

    @property
    def sample_freq_max(self):
        """Get the maximum sample frequency in Hz.

        Returns:
            float: Maximum sample frequency in Hz
        """
        return libtiepie.ScpGetSampleFrequencyMax(self._dev_handle)

    @property
    def sample_freq(self):
        """Get or set the current sample frequency in Hz."""
        return libtiepie.ScpGetSampleFrequency(self._dev_handle)

    @sample_freq.setter
    def sample_freq(self, value):
        libtiepie.ScpSetSampleFrequency(self._dev_handle, value)

    def verify_sample_freq(self, sample_frequency):
        """Verify a sample frequency without setting it in the hardware.

        Args:
            sample_frequency (float): The sample frequency to verify.
        Returns:
            float: The sample frequency the hardware would set. (The hardware
                   might not set the desired sample frequency due to clipping.)
        """
        return libtiepie.ScpVerifySampleFrequency(self._dev_handle,
                                                  sample_frequency)

    @property
    def record_length_max(self):
        """Get the maximum record length (number of samples).

        Returns:
            int: Maximum record length
        """
        return libtiepie.ScpGetRecordLengthMax(self._dev_handle)

    @property
    def record_length(self):
        """Get or set the current record length (number of samples)."""
        return libtiepie.ScpGetRecordLength(self._dev_handle)

    @record_length.setter
    def record_length(self, value):
        libtiepie.ScpSetRecordLength(self._dev_handle, value)

    def verify_record_length(self, record_length):
        """Verify a record length without setting it in the hardware.

        Args:
            record_length (int): The record length to verify.
        Returns:
            int: The record length the hardware would set. (The hardware
                 might not set the desired record length due to clipping.)
        """
        return libtiepie.ScpVerifyRecordLength(self._dev_handle,
                                               record_length)

    @property
    def pre_sample_ratio(self):
        """Get or set the pre sample ratio.

        The pre sample ratio is a float between 0 and 1 and defines how many
        samples should be recorded before the trigger point. To ensure all
        pre samples are recorded, set
        :py:attr:`handyscope.oscilloscope.Oscilloscope.trig_holdoff` to
        :py:attr:`handyscope.oscilloscope.Oscilloscope.TRIG_HOLDOFF_ALL_PRE_SAMPLES`
        or set :py:attr:`handyscope.oscilloscope.Oscilloscope.trig_holdoff` to equal
        or greater than
        :py:attr:`handyscope.oscilloscope.Oscilloscope.record_length` * pre_sample_ratio.
        """
        return libtiepie.ScpGetPreSampleRatio(self._dev_handle)

    @pre_sample_ratio.setter
    def pre_sample_ratio(self, value):
        libtiepie.ScpSetPreSampleRatio(self._dev_handle, value)

    @property
    def segment_cnt_max(self):
        """Get the maximum available segment count.

        Returns:
            int: Maximum available segment count
        """
        return libtiepie.ScpGetSegmentCountMax(self._dev_handle)

    @property
    def segment_cnt(self):
        """Get or set the current segment count."""
        return libtiepie.ScpGetSegmentCount(self._dev_handle)

    @segment_cnt.setter
    def segment_cnt(self, value):
        libtiepie.ScpSetSegmentCount(self._dev_handle, value)

    def verify_segment_cnt(self, segment_cnt):
        """Verify a segment count without setting it in the hardware.

        Args:
            segment_cnt (int): The segment count to verify.
        Returns:
            int: The segment count the hardware would set. (The hardware
                 might not set the desired segment count due to clipping.)
        """
        return libtiepie.ScpVerifySegmentCount(self._dev_handle,
                                               segment_cnt)

    @property
    def trig_timeout(self):
        """Get or set the trigger timeout in seconds.

        0 forces a trigger immediately after a measurement is started,
        -1 will wait infinitely for a trigger.
        """
        return libtiepie.ScpGetTriggerTimeOut(self._dev_handle)

    @trig_timeout.setter
    def trig_timeout(self, value):
        libtiepie.ScpSetTriggerTimeOut(self._dev_handle, value)

    def verify_trig_timeout(self, trig_timeout):
        """Verify a trigger timeout without setting it in the hardware.

        Args:
            trig_timeout (float): The trigger timeout to verify.
        Returns:
            float: The trigger timeout the hardware would set. (The hardware
                   might not set the desired trigger timeout due to clipping.)
        """
        return libtiepie.ScpVerifyTriggerTimeOut(self._dev_handle,
                                                 trig_timeout)

    @property
    def is_trig_delay_available(self):
        """Check if trigger delay is available.

        Returns:
            bool: True if trigger delay is available, False otherwise.
        """
        return libtiepie.ScpHasTriggerDelay(self._dev_handle) == 1

    @property
    def trig_delay_max(self):
        """Get the maximum available trigger delay in seconds.

        Returns:
            float: Maximum available trigger delay in seconds.
        """
        return libtiepie.ScpGetTriggerDelayMax(self._dev_handle)

    @property
    def trig_delay(self):
        """Get or set the current trigger delay in seconds."""
        return libtiepie.ScpGetTriggerDelay(self._dev_handle)

    @trig_delay.setter
    def trig_delay(self, value):
        libtiepie.ScpSetTriggerDelay(self._dev_handle, value)

    def verify_trig_delay(self, trig_delay):
        """Verify a trigger delay without setting it in the hardware.

        Args:
            trig_delay (float): The trigger delay to verify.
        Returns:
            float: The trigger delay the hardware would set. (The hardware
                   might not set the desired trigger delay due to clipping.)
        """
        return libtiepie.ScpVerifyTriggerDelay(self._dev_handle,
                                               trig_delay)

    @property
    def is_trig_holdoff_available(self):
        """Check if trigger holdoff is available.

        The trigger holdoff sets how many samples need to be recorded until
        a trigger can happen.

        Returns:
            bool: True if trigger holdoff is available, False otherwise.
        """
        return libtiepie.ScpHasTriggerHoldOff(self._dev_handle) == 1

    @property
    def trig_holdoff_max(self):
        """Get the maximum available trigger holdoff as number of samples.

        The trigger holdoff sets how many samples need to be recorded until
        a trigger can happen.

        Returns:
            float: Maximum available trigger holdoff as number of samples.
        """
        return libtiepie.ScpGetTriggerHoldOffCountMax(self._dev_handle)

    @property
    def trig_holdoff(self):
        """Get or set the current trigger holdoff as number of samples.

        The trigger holdoff sets how many samples need to be recorded until
        a trigger can happen.

        Use :py:attr:`handyscope.oscilloscope.Oscilloscope.TRIG_HOLDOFF_ALL_PRE_SAMPLES`
        to ensure all pre samples are recorded if pre_sample_ratio is set.
        """
        return libtiepie.ScpGetTriggerHoldOffCount(self._dev_handle)

    @trig_holdoff.setter
    def trig_holdoff(self, value):
        libtiepie.ScpSetTriggerHoldOffCount(self._dev_handle, value)

    @property
    def is_trig_available(self):
        """Check if trigger is available.

        Returns:
            bool: True if trigger is available, False otherwise.
        """
        return libtiepie.ScpHasTrigger(self._dev_handle) == 1

    @property
    def is_connection_test_available(self):
        """Check if connection test is available.

        Returns:
            bool: True if connection test is available, False otherwise.
        """
        return libtiepie.ScpHasConnectionTest(self._dev_handle) == 1

    def start_connection_test(self):
        """Start a connection test.

        Returns:
            bool: True if test started successfully, False otherwise.
        """
        return libtiepie.ScpStartConnectionTest(self._dev_handle) == 1

    @property
    def is_connection_test_completed(self):
        """Check if connection test is completed.

        Returns:
            bool: True if connection test is completed, False otherwise.
        """
        return libtiepie.ScpIsConnectionTestCompleted(self._dev_handle) == 1

    @property
    def connection_test_data(self):
        """Get connection test data.

        Returns:
            tuple: Tuple with the test result of each channel (key of
                   :py:attr:`handyscope.oscilloscope.Oscilloscope.CONNECTION_STATES`)
        """
        # Initialize uint8 array
        data = (ctypes.c_uint8 * self.channel_cnt)()

        # Write the actual data to the array
        libtiepie.ScpGetConnectionTestData(self._dev_handle, ctypes.byref(data),
                                           self.channel_cnt)

        # Convert to a normal python list
        data = list(data)

        # Evaluate
        data_evaluated = []
        for element in data:
            result = None

            # Try to look up the element
            for key, value in self.CONNECTION_STATES.items():
                if element == value:
                    result = key

            # If found append it to the output
            if result is not None:
                data_evaluated.append(result)
            # Else throw an exception
            else:
                raise ValueError("Unknown connection state: %d" % element)

        return tuple(data_evaluated)

    def test_connection(self):
        """Perform a connection test.

        Returns:
            tuple: Tuple with the test result of each channel (key of
                   :py:attr:`handyscope.oscilloscope.Oscilloscope.CONNECTION_STATES`)
        """
        if self.is_connection_test_available:
            res = self.start_connection_test()
            if res is False:
                raise IOError("Connection test could not be started.")

            while not self.is_connection_test_completed:
                time.sleep(0.1)

            return self.connection_test_data
        else:
            return None

    def measure(self, safe=True):
        """Perform a single shot measurement.

        Utility function which starts a measurement. When measurement data is
        ready, it is retrieved and returned.

        Args:
            safe(bool): This is only relevant when using presamples.
                        If at the start of the measurement the trigger occurs
                        before even enough samples have been recorded for
                        the pre samples, the desired record length will not
                        match to the actual returned data length. Make sure to
                        set the trig_holdoff accordingly. If set to True, an
                        exception will be raised if not all samples could have
                        been collected. If set to False or no trig_holdoff is
                        available warning will be raised.


        Returns:
            list: List with entries for each channel. An entry contains None,
                  if the channel is disabled, otherwise a list of samples.
        """
        if self.measure_mode == "block" and self.pre_sample_ratio > 0:
            if self.is_trig_holdoff_available and self.trig_holdoff < self.pre_sample_ratio * self.record_length:
                warnings.warn("trig_holdoff is not set to record all pre"
                              " samples. This may lead to varying signal "
                              "lengths due to triggers occuring before "
                              "enough samples have been stored."
                              " Please set trig_holdoff to "
                              "TRIG_HOLDOFF_ALL_PRE_SAMPLES to get a "
                              "predictable time vector.", UserWarning)
        # Start measurement
        self.start()

        # Wait until measurement is finished
        while not self.is_data_ready:
            time.sleep(0.1)

        if self.measure_mode == "block":
            if self.valid_pre_sample_cnt < self.record_length * self.pre_sample_ratio:
                if safe and self.is_trig_holdoff_available:
                    raise ValueError("Not all presamples have been collected")
                else:
                    warnings.warn("Not all presamples have been collected",
                                  UserWarning)
        # Get data
        data = self.retrieve()

        return data

    @property
    def time_vector(self):
        """Get a time vector according to the current oscilloscope settings.

        Returns:
            :class:`numpy.ndarray`: Time vector
        """
        time_vec = np.linspace(0, 1 / self.sample_freq * self.record_length,
                               num=self.record_length, endpoint=False)
        if self.measure_mode == "block":
            trig_idx = int(self.pre_sample_ratio * len(time_vec))
            return time_vec - time_vec[trig_idx]
        else:
            return time_vec



from handyscope.device import Device
from handyscope.deviceList import DeviceList
from handyscope.oscilloscope import Oscilloscope
from handyscope.generator import Generator
from handyscope.i2cHost import I2CHost
import handyscope.deviceList
import pytest


def pytest_addoption(parser):
    parser.addoption("--product_id", action="store", default="HS5")


def pytest_generate_tests(metafunc):
    option_value = metafunc.config.option.product_id
    if 'product_id' in metafunc.fixturenames and option_value is not None:
        metafunc.parametrize("product_id", [option_value], scope="module")


@pytest.fixture(scope="module")
def dev_list():
    return handyscope.deviceList.device_list


@pytest.fixture(scope="module", params=[key for key in DeviceList.DEVICE_TYPES])
def device(product_id, request):
    dev_instance = Device(product_id, "product id", request.param)
    yield dev_instance
    dev_instance.close()


@pytest.fixture(scope="module")
def osc(product_id):
    osc_instance = Oscilloscope(product_id)
    yield osc_instance
    osc_instance.close()


@pytest.fixture(scope="function")
def default_osc(osc):
    if osc.is_running:
        osc.stop()
    osc.measure_mode = "block"
    osc.clock_source = "internal"
    osc.auto_resolution = "disabled"
    osc.resolution = 14
    osc.record_length = 5000
    osc.segment_cnt = 1
    osc.trig_delay = 0.0
    osc.trig_holdoff = 0

    osc.channels[0].is_enabled = True
    for channel in osc.channels[1:]:
        channel.is_enabled = False

    return osc


@pytest.fixture(scope="module")
def gen(product_id):
    gen_instance = Generator(product_id)
    yield gen_instance
    gen_instance.close()


@pytest.fixture(scope="function")
def default_gen_sine(gen):
    gen.is_amplitude_autorange = True
    gen.signal_type = "sine"
    gen.mode = "continuous"
    gen.offset = 0.0
    gen.amplitude = 1.0
    gen.stop()

    return gen


@pytest.fixture(scope="function")
def default_gen_pulse(gen):
    gen.is_amplitude_autorange = True
    gen.signal_type = "pulse"
    gen.mode = "continuous"
    gen.offset = 0.0
    gen.amplitude = 1.0
    gen.freq = 1000.0

    return gen


@pytest.fixture(scope="function")
def default_gen_arb(gen):
    gen.is_amplitude_autorange = True
    gen.signal_type = "arbitrary"
    gen.mode = "continuous"
    gen.offset = 0.0
    gen.amplitude = 1.0

    return gen


@pytest.fixture(scope="function")
def default_gen_burst(gen):
    gen.is_amplitude_autorange = True
    gen.signal_type = "sine"
    gen.mode = "burst count"
    gen.offset = 0.0
    gen.amplitude = 1.0

    return gen


@pytest.fixture(scope="function")
def default_gen_burst_sample(gen):
    gen.is_amplitude_autorange = True
    gen.signal_type = "arbitrary"
    gen.freq_mode = "sample"
    gen.arb_data([0.0, 1.0, 2.0, 3.0, 4.0])
    gen.mode = "burst sample count"
    gen.offset = 0.0
    gen.amplitude = 1.0

    return gen


@pytest.fixture(scope="function")
def default_gen_burst_segment(gen):
    gen.is_amplitude_autorange = True
    gen.signal_type = "arbitrary"
    gen.arb_data([0.0, 1.0, 2.0, 3.0, 4.0])
    gen.mode = "burst segment count"
    gen.offset = 0.0
    gen.amplitude = 1.0

    return gen


@pytest.fixture(scope="module")
def i2c(product_id):
    i2c_instance = I2CHost(product_id)
    yield i2c_instance
    i2c_instance.close()



from handyscope.library import libtiepie
import ctypes


class TriggerOutput:
    TRIGGER_EVENTS = {"unknown":              0,
                      "Generator start":      1,
                      "Generator stop":       2,
                      "Generator new period": 4,
                      "Oscilloscope running": 8,
                      "Oscilloscope triggered": 16,
                      "manual": 32,
                      }

    # see api doc, macro "TRIGGER_IO_ID"
    TRIGGER_IDS = {"EXT 1": 0 << 24 | 3 << 20 | 1 << 8 | 0,
                   "EXT 2": 0 << 24 | 3 << 20 | 2 << 8 | 0,
                   "EXT 3": 0 << 24 | 3 << 20 | 3 << 8 | 0}

    def __init__(self, dev_handle, trig_out_idx):
        self._dev_handle = dev_handle
        self._idx = trig_out_idx

    @property
    def is_enabled(self):
        return libtiepie.DevTrOutGetEnabled(self._dev_handle, self._idx) == 1

    @is_enabled.setter
    def is_enabled(self, value):
        libtiepie.DevTrOutSetEnabled(self._dev_handle, self._idx, value)

    @property
    def trigger_id(self):
        raw_id = libtiepie.DevTrOutGetId(self._dev_handle, self._idx)
        for key in self.TRIGGER_IDS:
            if self.TRIGGER_IDS[key] == raw_id:
                return key

        raise ValueError("Unknown trigger output id %d" % raw_id)

    @property
    def name(self):
        # get length of name string
        str_len = libtiepie.DevTrOutGetName(
            self._dev_handle, self._idx, None, 0)

        # initialize mutable string buffer
        str_buffer = ctypes.create_string_buffer(str_len)

        # write the actual name to the buffer
        libtiepie.DevTrOutGetName(
            self._dev_handle, self._idx, str_buffer, str_len)

        # convert to a normal python string
        name = str_buffer.value.decode('utf-8')

        return name

    @property
    def events_available(self):
        raw_events = libtiepie.DevTrOutGetEvents(self._dev_handle, self._idx)
        _events = []

        # if no trigger evvents are available, return unknown
        if raw_events == self.TRIGGER_EVENTS["unknown"]:
            _events.append("unknown")
        # else do a detailed analysis...
        else:
            # ... by iterating over every possible kind...
            for key in self.TRIGGER_EVENTS:
                # ... and ignoring "unknown" (already handled above)
                if key == "unknown":
                    pass
                elif raw_events & self.TRIGGER_EVENTS[key] == self.TRIGGER_EVENTS[key]:
                    _events.append(key)

        return _events

    @property
    def event(self):
        raw_event = libtiepie.DevTrOutGetEvent(self._dev_handle, self._idx)
        for key in self.TRIGGER_EVENTS:
            if raw_event == self.TRIGGER_EVENTS[key]:
                return key

        raise ValueError("Unknown trigger output event: %d" % raw_event)

    @event.setter
    def event(self, value):
        libtiepie.DevTrOutSetEvent(
            self._dev_handle, self._idx, self.TRIGGER_EVENTS[value])

    def force_trig(self):
        return libtiepie.DevTrOutTrigger(self._dev_handle, self._idx) == 1



def test_device_cnt(dev_list):
    assert type(dev_list.device_cnt) is int
    assert dev_list.device_cnt > 0


def test_can_open_device(dev_list):
    assert dev_list.can_open_device(0) is True


def test_get_firmware_version(dev_list):
    assert type(dev_list.get_firmware_version(0)) is str


def test_get_recommended_firmware_version(dev_list):
    assert type(dev_list.get_firmware_version(0)) is str


def test_get_device_name(dev_list):
    assert type(dev_list.get_device_name(0)) is str


def test_get_device_name_short(dev_list):
    assert type(dev_list.get_device_name_short(0)) is str


def test_get_device_name_shortest(dev_list):
    assert type(dev_list.get_device_name_shortest(0)) is str


def test_get_device_serial_no(dev_list):
    assert type(dev_list.get_device_serial_no(0)) is int


def test_get_device_types(dev_list):
    assert type(dev_list.get_device_types(0)) is dict
    for key, value in dev_list.get_device_types(0).items():
        assert key in dev_list.DEVICE_TYPES
        assert type(value) is bool


def test_get_overview(dev_list):
    assert type(dev_list.get_overview()) is tuple
    for element in dev_list.get_overview():
        assert type(element) is dict
        assert element["Index"] in range(dev_list.device_cnt)
        assert element["Name"] == dev_list.get_device_name_short(element["Index"])
        assert element["SerNo"] == dev_list.get_device_serial_no(element["Index"])
        assert element["DevTypes"] == dev_list.get_device_types(element["Index"])


def test_get_overview_str(dev_list):
    assert type(dev_list.get_overview_str()) is str


def test__str__(dev_list):
    assert str(dev_list) == dev_list.get_overview_str()


# Opening a device is tested when instantiating Oscilloscope, Generator and I2CHost
